import ClaraScene from './clara-scene';
import ClaraAnimation from './clara-animation';
import ClaraEvents from './clara-events';
import ColorHelper from './color-helper';
import ClaraConstants from './clara-constants';
import ClaraSnapshot from './clara-snapshot';

export default class ClaraFacade {

  sceneContainerId = null;
  claraApi = null;

  events = null;
  scene = null;
  animation = null;
  snapshot = null;

  lookAtNodeId = null;
  outOfFocusMaterialId = null;
  initialMeshMaterials = null;

  initializeScene = sceneContainerId => {
    this.sceneContainerId = sceneContainerId;
    this.claraApi = claraplayer(this.sceneContainerId); // eslint-disable-line no-undef
    this.claraApi.player.displayThumbnail(false);
    this.claraApi.selection.setHighlighting(true);
    this.claraApi.commands.updateCommand('select', {enabled: false});

    this.events = new ClaraEvents(this.claraApi, this.sceneContainerId);
    this.scene = new ClaraScene(this.claraApi);
    this.animation = new ClaraAnimation(this.claraApi);
    this.snapshot = new ClaraSnapshot(this.claraApi);

    this._registerEvents();
  };

  _registerEvents = () => {
    this.events.registerOnPreloaded(this._onClaraScenePreloaded);
    this.events.registerOnRendered(this._onClaraSceneRendered);
    this.events.registerOnHomeButtonPressed(this.animation.resetAllAnimations);
  };

  _onClaraScenePreloaded = () => {
    const baseColor = { r: 1, g: 0.5, b: 0.5 };
    this.outOfFocusMaterialId = this.scene.createMaterial('_Out_of_focus_MAT_', { opacityFactor: 0.5, baseColor });
    const meshMaterialsReference = {
      type: ClaraConstants.POLYMESH,
      plug: ClaraConstants.MATERIAL,
      property: ClaraConstants.REFERENCE
    };
    this.initialMeshMaterials = this.claraApi.scene.getAll(meshMaterialsReference);
  };

  _onClaraSceneRendered = () => {
    const root = this.scene.findRootModelOrNull();
    const lookAtNode = this.claraApi.scene.addNode({
      name: 'LOOK_AT_NODE',
      type: ClaraConstants.NULL,
      parent: root
    });
    this.lookAtNodeId = this.scene.getId(lookAtNode);
  };

  renderScene = claraSceneId => {
    this.claraApi.sceneIO.fetchAndUse(claraSceneId, null, {waitForPublish: true});
  };

  renderPublishedScene = (claraSceneId, hash) => {
    this.claraApi.sceneIO.fetchAndUse(claraSceneId, hash);
  };

  animateToMainCamera = () => {
    const mainCameraId = this.claraApi.scene.find({ type: ClaraConstants.CAMERA, name: ClaraConstants.CAMERA });
    const mainCameraTransform = this.scene.getTransform(mainCameraId);
    const easingTranslation = {
      x: mainCameraTransform.position.z * -2,
      y: mainCameraTransform.position.y,
      z: mainCameraTransform.position.x
    };
    const root = this.claraApi.scene.find({ name: 'Objects' });
    const easingTarget = this.scene.getId(this.claraApi.scene.addNode({name: 'easingTarget', type: ClaraConstants.NULL, parent: root}));
    const easingCameraId = this.scene.getId(this.claraApi.scene.addNode({
      name: 'easingCamera',
      type: ClaraConstants.CAMERA,
      parent: root,
      plugs: {
        Transform: [
          [ClaraConstants.TRANSFORM, {
            translation: easingTranslation
          }],
          ['LookAt', {target: easingTarget}]
        ]
      }
    }));
    this.claraApi.player.useCamera(easingCameraId);
    return this.claraApi.player.animateCameraTo(mainCameraId, 2000);
  };

  disableTools = tools => {
    tools.forEach(tool => this.disableTool(tool));
  };

  hideTools = tools => {
    tools.forEach(tool => this.hideTool(tool));
  };

  disableTool = tool => {
    this.claraApi.player.removeTool(tool);
  };

  hideTool = tool => {
    this.claraApi.player.hideTool(tool);
  };

  hideToolbar = () => {
    Object.values(ClaraConstants.toolbar).forEach(tool => this.hideTool(tool));
  };

  enableHelpTool = onShowHelp => {
    this.claraApi.commands.addCommand({
      options: { icon: 'help' },
      display: { playerTools: true },
      label: 'help',
      execute: () => {
        this.claraApi.player.exitFullscreen();
        onShowHelp();
      }
    });
  };

  setActiveColor = rgbColor => {
    this.claraApi.player.setActiveColor(rgbColor);
  };

  setHoverColor = rgbColor => {
    this.claraApi.player.setHoverColor(rgbColor);
  };

  showMesh = name => {
    const nodeId = this.scene.findMeshOrNullByName(name);

    if (!nodeId) {
      console.error(`Cannot find node named: ${name}`);
      return;
    }

    this.hideAllSiblingMeshes(nodeId);
  };

  hideMesh = name => {
    const nodeId = this.scene.findMeshOrNullByName(name);
    this.setVisibilityRecursively(nodeId, false);
  };

  hideAllSiblingMeshes = nodeId => {
    const parentId = this.scene.findParent(nodeId);

    const siblings = this.claraApi.scene.filter({
      from: { id: parentId },
      shallow: true
    });

    siblings.forEach(siblingId => {
      this.setVisibilityRecursively(siblingId, false);
    });
    this.setVisibilityRecursively(nodeId, true);
  };

  spawnAnnotation = (childName, name) => {
    const childId = this.scene.findMeshOrNullByName(childName);

    if (!childId) {
      console.error(`Could not find node with name "${childName}" in order to place a hotspot!`);
      return null;
    }

    const annotation = this.claraApi.sceneGraph.addNode({
      name,
      type: ClaraConstants.ANNOTATION,
      parent: childId,
      plugs: {
        Annotation: [[ClaraConstants.ANNOTATION, {
          limitVisilibility: true,
          directionMode: 'awayFromOrigin',
          directionTarget: this.claraApi.scene.find({ type: ClaraConstants.MODEL })
        }]],
        Transform: [[ClaraConstants.TRANSFORM, {}]],
        Properties: [['Default', {}]]
      }
    });
    return this.scene.getId(annotation);
  };

  setVisibilityRecursively = (nodeId, isVisible) => {
    this.claraApi.scene.setAll({
      from: { id: nodeId },
      plug: 'Properties',
      property: ClaraConstants.VISIBLE,
      includeParent: true
    }, isVisible);
  };

  overrideClaraAnnotations = updateFunctionToCall => {
    this.claraApi.annotations.useCustomAnnotationFunction(updateFunctionToCall);
  };

  changeColor = (materialName, r, g, b) => {
    const materialId = this.scene.findMaterial(materialName);
    if (!materialId) {
      console.error(`Cannot find material named: ${materialName}`);
      return;
    }

    const baseColor = this.claraApi.scene.find({
      id: materialId,
      plug: ClaraConstants.MATERIAL,
      property: ClaraConstants.BASE_COLOR
    });
    this.claraApi.scene.set(baseColor, this.createColorObjectForClara(r, g, b));
  };

  changeBaseTexture = (materialName, textureName) => {
    const materialId = this.scene.findMaterial(materialName);
    if (!materialId) {
      console.error(`Cannot find material named: ${materialName}`);
      return;
    }

    const textureId = this.scene.findTexture(textureName);
    if (!textureId) {
      console.error(`Cannot find texture named: ${textureName}`);
      return;
    }

    const materialBaseMap = { id: materialId, plug: ClaraConstants.MATERIAL, property: ClaraConstants.BASE_MAP };
    this.claraApi.scene.set(materialBaseMap, textureId);
  };

  createColorObjectForClara = (r, g, b) => {
    return ColorHelper.convertColorGammaToLinear(ColorHelper.createColorObject0to1(r, g, b));
  };

  convertPositionOffsetUnityToClara = offset => {
    const result = {...offset};
    result.x *= -100;
    result.y *= 100;
    result.z *= 100;
    return result;
  };

  swapMaterialByMeshName = (meshName, materialName) => {
    const meshNodeId = this.claraApi.scene.find({ type: ClaraConstants.POLYMESH, name: meshName });
    if (!meshNodeId) {
      console.error(`Could not find a mesh with name "${meshName}" in order to swap the material.`);
      return;
    }

    const materialId = this.scene.findMaterial(materialName);
    if (!materialId) {
      console.error(`Could not find material with name "${materialName}' in order to swap the material.`);
      return;
    }
    this.scene.setMaterial(meshNodeId, materialId);
  };

  rotateAroundWorldAxis = (nodeId, rotation) => {
    const THREE = this.claraApi.deps.THREE;
    const oldParentId = this.scene.findParent(nodeId);
    const nodeTransform = this.claraApi.scene.getWorldTransform(nodeId);

    const eulerRotation = new THREE.Euler(rotation.x, rotation.y, rotation.z, 'ZYX');
    const newTransformMatrix = new THREE.Matrix4().makeRotationFromEuler(eulerRotation).multiply(nodeTransform);
    const parentTransform = this.claraApi.scene.getWorldTransform(oldParentId);

    let newRelativeTransform;
    if (parentTransform) {
      const parentWorldInverse = new THREE.Matrix4().getInverse(this.claraApi.scene.getWorldTransform(oldParentId));
      newRelativeTransform = new THREE.Matrix4().multiplyMatrices(parentWorldInverse, newTransformMatrix);
    } else {
      newRelativeTransform = newTransformMatrix;
    }

    const newNodeTranslate = new THREE.Vector3();
    const newNodeRotate = new THREE.Quaternion();
    const newNodeScale = new THREE.Vector3();
    newRelativeTransform.decompose(newNodeTranslate, newNodeRotate, newNodeScale);
    const rotateVec = new THREE.Euler().setFromQuaternion(newNodeRotate, 'ZYX').toVector3().multiplyScalar(180.0 / Math.PI);

    this.scene.setTransform(nodeId, newNodeTranslate, rotateVec, newNodeScale);
  };

  swapMaterials = (oldMaterialName, newMaterialName) => {
    const newMaterialId = this.scene.findMaterial(newMaterialName);
    if (!newMaterialId) {
      console.error(`Could not find material with name "${newMaterialName}" to swap`);
      return;
    }

    const materialPaths = this.scene.findMeshesWithMaterialByName(oldMaterialName);
    if (materialPaths.length === 0) {
      console.error(`Found no meshes that use material "${oldMaterialName}" to swap`);
      return;
    }

    this.claraApi.scene.setAll(materialPaths, newMaterialId);
  };

  addExecuteToolbarButton = (iconName, onClickedCallback) => {
    return this.claraApi.commands.addCommand({
      options: { icon: iconName },
      display: { playerTools: true },
      execute: onClickedCallback
    });
  };

  addToggleToolbarButton = (iconName, onActivateCallback, onDeactivateCallback) => {
    return this.claraApi.commands.addCommand({
      options: { icon: iconName },
      display: { playerTools: true },
      activate: onActivateCallback,
      deactivate: onDeactivateCallback
    });
  };

  flyTo = (nodeName, zoomLevel) => {
    const nodeId = this.scene.findMeshOrNullByName(nodeName);
    if (nodeId) {
      const zoomLevelValue = !zoomLevel ? this.getAutoCalculatedZoomLevel(nodeId) : zoomLevel;
      this.claraApi.selection.selectNode(nodeId);
      const nodeCenter = this.getCenter(nodeId);
      const lookAtCameraId = this.createSecondaryLookAtCamera(nodeCenter, zoomLevelValue);
      this.claraApi.player.animateCameraTo(lookAtCameraId, 2000)
        .then(() => {
          this.claraApi.sceneGraph.deleteNode(lookAtCameraId);
        });
    } else {
      console.error(`Could not find node with name "${nodeName}" to fly to.`);
    }
  };

  focusOn = (nodeName, zoomLevel) => {
    this.unfocusAll();
    this.flyTo(nodeName, zoomLevel);
    const nodeId = this.scene.findMeshOrNullByName(nodeName);
    if (!nodeId) {
      console.error(`Could not find node with name "${nodeName}" to focus on.`);
      return;
    }
    this.claraApi.selection.selectNode(nodeId);
    const meshMaterialReferencePaths = this.claraApi.scene.filter({
      type: ClaraConstants.POLYMESH,
      plug: ClaraConstants.MATERIAL,
      property: ClaraConstants.REFERENCE
    });
    const focusedNodes = this.claraApi.scene.filter({
      from: {id: nodeId},
      includeParent: true
    });
    const contextNodes = this.claraApi.scene.filter({
      from: {name: 'Context'},
      type: ClaraConstants.POLYMESH,
      includeParent: true
    });
    const nonSelectedMeshes = meshMaterialReferencePaths.filter(path => {
      return !focusedNodes.includes(path[0]) && !contextNodes.includes(path[0]);
    });
    // Apply the newly created material to all unfocused meshes.
    this.claraApi.scene.setAll(nonSelectedMeshes, this.outOfFocusMaterialId);
  };

  unfocusAll = () => {
    Object.keys(this.initialMeshMaterials).forEach(meshId => {
      this.scene.setMaterial(meshId, this.initialMeshMaterials[meshId]);
    });
    this.claraApi.selection.deselectAll();
  };

  getAutoCalculatedZoomLevel = nodeId => {
    const boundingLength = this.getAverageBoundingBoxLength(nodeId);
    return 0.0995359 + (2.906155 * boundingLength) + (-0.7061069 * boundingLength * boundingLength);
  }

  getAverageBoundingBoxLength = nodeId => {
    const boundingBox = this.claraApi.scene.getNodeBoundingBox(nodeId);
    const lengthX = boundingBox.max.x - boundingBox.min.x;
    const lengthY = boundingBox.max.y - boundingBox.min.y;
    const lengthZ = boundingBox.max.z - boundingBox.min.z;
    return (lengthX + lengthY + lengthZ) / 3;
  }

  getCenter = nodeId => {
    const THREE = this.claraApi.deps.THREE;
    const boundingBox = this.claraApi.scene.getNodeBoundingBox(nodeId);
    const center = new THREE.Vector3();
    center.x = (boundingBox.min.x + boundingBox.max.x) / 2;
    center.y = (boundingBox.min.y + boundingBox.max.y) / 2;
    center.z = (boundingBox.min.z + boundingBox.max.z) / 2;
    return center;
  }

  createSecondaryLookAtCamera = (partCenter, zoomLevel) => {
    const root = this.scene.findRootModelOrNull();
    const lookAtNodeTranslation = {
      id: this.lookAtNodeId,
      plug: ClaraConstants.TRANSFORM,
      property: ClaraConstants.TRANSLATION
    };
    this.claraApi.scene.set(lookAtNodeTranslation, partCenter);

    const partTranslationInWorld = this.getTranslationInWorld(this.lookAtNodeId);
    const originToPartVectorNormalized = partTranslationInWorld.clone().normalize();
    const cameraVector = originToPartVectorNormalized.multiplyScalar(zoomLevel);
    const cameraPosition = partTranslationInWorld.add(cameraVector);

    const camera = this.claraApi.scene.addNode({
      name: 'secondaryCamera',
      type: ClaraConstants.CAMERA,
      parent: root,
      plugs: {
        Transform: [
          [ClaraConstants.TRANSFORM, {
            translation: cameraPosition
          }],
          ['LookAt', {
            target: this.lookAtNodeId
          }]
        ],
        Camera: [
          [ClaraConstants.CAMERA, {
            nearClip: 0.01
          }]
        ]
      }
    });
    return this.scene.getId(camera);
  };

  getTranslationInWorld = nodeId => {
    const THREE = this.claraApi.deps.THREE;
    const partTransformMatrix = this.claraApi.scene.getWorldTransform(nodeId);
    const translation = new THREE.Vector3();
    translation.setFromMatrixPosition(partTransformMatrix);
    return translation;
  };

  setMeshQuality = (level = 0) => {
    const meshQualityPaths = this.claraApi.scene.filter({type: ClaraConstants.POLYMESH, plug: ClaraConstants.POLYMESH, property: `level${level}`});
    const meshIdsToQualityIds = this.claraApi.scene.getAll(meshQualityPaths);
    Object.entries(meshIdsToQualityIds).forEach(([meshId, qualityLevelId]) => {
      const meshBinary = { id: meshId, plug: ClaraConstants.POLYMESH, property: ClaraConstants.BINARY };
      this.claraApi.scene.set(meshBinary, qualityLevelId);
    });
  };

  getMeshQuality = () => {
    const meshId = this.claraApi.scene.find({type: 'PolyMesh'});
    const currentQualityLevelId = this.claraApi.scene.get({id: meshId, plug: 'PolyMesh', property: 'binary'});
    for (let level = 0; level <= 4; level += 1) {
      const qualityLevelId = this.claraApi.scene.get({ id: meshId, plug: 'PolyMesh', property: `level${level}` });
      if (qualityLevelId === currentQualityLevelId) {
        return level;
      }
    }
    return -1;
  };
}
